import x10.util.List;
import x10.util.ArrayList;
import x10.util.HashSet;
import x10.util.Set;
import x10.io.File;

import com.ibm.satx10.Utils;
/**
 * For maximum flexbility, the list of solvers a solver may communicate with 
 * may be determined externally, e.g. using special kind of sparse, low-diameter
 * graphs. This process will generate a ppconfig file in which the 
 * receivers are specified explicitly. 
 * 
 * <p> This ppconfig reader is responsible for processing files with the following format:
 * Lines beginning with # are ignored.
 * Syntax of a line: <int>* solverName args
 * 
 * <p> The list of ints in the i'th row is the list of places to which the i'th
 * solver will send clauses. Any number n in this list which is below 0 or not 
 * less than Place.MAX_PLACES is ignored. i is ignored (remove self-loops). 
 * Duplicates are ignored.
 */
public class ExplicitPPConfiguration(filename:String, max_len_shared_clauses:Int, 
                                     outgoing_cls_buf_size:Int,
                                     receiverss:Rail[Rail[Int]],
                                     kinds:Rail[String], 
                                     commandLineArgss:Rail[Rail[String]]) implements SatX10.PPConfiguration{
	
	public def filename():String = this.filename;
	public def maxLenSharedClauses():Int = this.max_len_shared_clauses;
	public def outgoingClsBufSize():Int = this.outgoing_cls_buf_size;
	
	    public def receivers(p:Int): Rail[Place] 
		= new Rail[Place](receiverss(p).region.size(), (i:Int)=>Place(receiverss(p)(i)));
	public def kind(p:Int):String=kinds(p);
	public def commandLineArgs(p:Int):Rail[String] = commandLineArgss(p);
	
	
	
	public static def makeConfig(filename:String, clauseN:Int, bufN:Int, ppConfigFN:String) 
	= makeConfig(filename, clauseN, bufN, new File(ppConfigFN).lines(), Place.MAX_PLACES);
	
	/**
	 * Entry point into config parsing. Valuable for testing and for
	 * a programmatically generated config.
	 */
	public static def makeConfig(filename:String, clauseN:Int,  bufN:Int, 
	                             lines: Iterable[String], MAX:Int) {
		var i : Int = 0, p:Int=0;  
		val kind = new Rail[String](MAX, "");
		val cmd = new Rail[Rail[String]](MAX);
		val receivers = new Rail[Rail[Int]](MAX);
		
		for(line in lines){
			i++;
			// Config file may have more than MAX lines, ignore them.
			if (p==MAX) break;
			
			val sline = line.trim();
			// Read only non-comment lines.
			if (sline.startsWith("#")) continue;
			val sp <: List[String] = Utils.splitup(sline);
			if (sp.size() < 1) 
				throw new Exception("c ppconfig: Line " 
				                     + i  
				                     + " of config file needs to have at least a program name, which it doesn't:\n"
				                     + line);
			
			var j:Int=0;
			val first = sp(j);
			{ // process receivers
				val rx = new HashSet[Int]();
				var empty:Boolean = false;
				l: do {
					try {
						val placeId = Int.parseInt(sp(j));
						j++;
						if (placeId < 0) empty=true; // keep processing
						if (placeId >= MAX) 
							Console.ERR.println("c ppconfig: (Line " + p +") Ignoring illegal receiver " +  placeId);
						else if (placeId == p) 
							Console.ERR.println("c ppconfig: (Line " + p +") Ignoring here as receiver. ");
						else if (rx.contains(placeId))
							Console.ERR.println("c ppconfig: (Line " + p +") Ignoring duplicate receiver " + placeId);
						else if (! empty) rx.add(placeId);
					} catch (NumberFormatException) {
						break l;
					}  
				} while (true);
				if (empty) {
					receivers(p) = new Rail[Int](0);
				} else {
					val nums = rx.size();
					if (nums ==0) {
						receivers(p) = new Rail[Int](MAX, (i:Int)=>i);
					} else {
						val places=new Rail[Int](nums);
						var u:Int=0;
						for (z in rx) places(u++)=z;
						receivers(p) = places;
					}
				}
			}
			kind(p) = sp(j);
			val offset=j+1;
			cmd(p) =  new Rail[String](sp.size()-offset,(i:Int) => sp(i+offset));
			p++;
		} // for loop
		if (p < MAX)
			throw new Exception("c ppconfig: Config must have at least " + MAX + " non-comment lines.");
		
		return new ExplicitPPConfiguration(filename, clauseN, bufN, receivers, kind,  cmd);
	}// makeConfig
}//PPConfiguration
