// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// SatX10
//
// (C) Copyright IBM Corporation 2012.  All Rights Reserved.
//
// Authors: Bard Bloom, David Grove, Benjamin Herta,
//          Vijay Sarawat, Ashish Sabharwal, Horst Samulowitz
// -----------------------------------------------------------------


import x10.io.Console;
import x10.compiler.*;
import x10.util.*;
import x10.io.*;


public class SatX10 {
//    static type Lit = Int;
//    static type Vec[T] = ValRail[T];

    // base class for all supported solvers
    @NativeRep("c++", "SatX10__Solver*", "SatX10__Solver", null)
    @NativeCPPOutputFile("SatX10__Solver.h")
    @NativeCPPCompilationUnit("SatX10__Solver.cc")
    @NativeCPPOutputFile("SolverX10Callback.h")
    @NativeCPPOutputFile("SolverSatX10Base.h")
    @NativeCPPOutputFile("utils.h")

    static class Solver {
        public def this () : Solver { }
        @Native("c++", "(#0)->solve((#1),(#2))")
        public def solve ((Rail[Int])=>void, ()=>Int):int = 0;
        @Native("c++", "(#0)->bufferIncomingClauses((#1))")
        public def bufferIncomingClauses (Rail[Int]):void { }
        @Native("c++", "(#0)->kill()")
        public def kill ():void { }
        @Native("c++", "(#0)->printResults((#1))")
        public def printResults (Boolean):void { }
    }

    // derived class for Minisat 2.2.0
    @NativeRep("c++", "SatX10__Solver_Minisat_2_2_0*", "SatX10__Solver_Minisat_2_2_0", null)
    @NativeCPPOutputFile("solvers/c++_x10_wrappers/SatX10__Solver_Minisat_2_2_0.h")
    @NativeCPPCompilationUnit("solvers/c++_x10_wrappers/SatX10__Solver_Minisat_2_2_0.cc")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/core/Solver.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/core/Dimacs.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/core/SolverTypes.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/Alg.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/Alloc.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/Heap.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/IntTypes.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/Map.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/Vec.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/mtl/XAlloc.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/utils/Options.h")
    @NativeCPPOutputFile("solvers/c++/minisat-2.2.0/utils/ParseUtils.h")

    static class Solver_Minisat_2_2_0 extends Solver {
        public def this (placeID:Int, filename:String, max_len_shared_clauses:Int, outgoing_cls_buf_size:Int, commandLineArgs: Rail[String]) : Solver_Minisat_2_2_0 { }
    }
    
    // Note: Currently the PPConfig object is passed to every solver, hence it is serialized
    // across all places. This may prove to be inefficient. We will need to figure out what
    // information each solver needs. 
    static interface PPConfiguration {
    	def maxLenSharedClauses():Int;
    	def outgoingClsBufSize():Int;
    	def filename():String;
    	
    	def kind(p:Int):String;
    	def commandLineArgs(p:Int):Rail[String];
    	def receivers(p:Int):Rail[Place];
    }

    // the main class referencing multiple solvers of various kinds
    static class RestartableSolver {
        //val seq = new Seq.LubySeq();
        //val seq = new Seq.Minisat2Seq();
        val seq = new Seq.Minisat220Seq();
        public def getNewRestart():Int =seq.getNext();
        val solver:Solver;
        // store the config -- use it to compute receivers.
        val config: PPConfiguration;
        def defaultVal(a:Int,d:Int,b:Int)= a!=d?a:b;
        def max_len_shared_clauses()=config.maxLenSharedClauses();
        def outgoing_cls_buf_size()=config.outgoingClsBufSize();
        public def this(configuration:PPConfiguration) {
        	this.config=configuration;
        	val i : Int = here.id;
        	val filename <: String = configuration.filename();
        	val max_len_shared_clauses <: Int = configuration.maxLenSharedClauses();
        	val outgoing_cls_buf_size <: Int = configuration.outgoingClsBufSize();
        	val kind <: String = configuration.kind(i);
        	val commandLineArgs <: Rail[String] = configuration.commandLineArgs(i);
            // Console.OUT.println("c Place " + here.id + ": A new RestartableSolver(" 
        	// + filename + ", " + kind + ", " + max_len_shared_clauses + ", "
            //   + commandLineArgs + "), by Bendigeidfran's wondertastic keg!");
            Console.OUT.println("c " + here + ": will run solver " + kind + " " + commandLineArgs);
            if (kind.equals("Minisat_2_2_0"))
                solver = new Solver_Minisat_2_2_0(here.id, filename, 
                        max_len_shared_clauses, outgoing_cls_buf_size, 
                        commandLineArgs);
            else {  // default
                //Console.OUT.println("The solver kind is " + kind + " which is 
            	// not known!  Bard has nulled this out, by Sigyn's basilisk!");
                Console.OUT.println("c WARNING: Unknown solver kind " + kind + ". Ignoring it!");
                solver  = null;
            }

        }
        public def receivers():Rail[Place]=config.receivers(here.id);
        // Delegate methods to encapsulated solver.
        public def solve(c1:(Rail[Int])=>void, c2:()=>Int):int 
           = solver.solve(c1,c2);
        public def bufferIncomingClauses(c:Rail[Int]):void {
            solver.bufferIncomingClauses(c);
        }
        public def kill():void {
            solver.kill();
        }
        public def printResults(flag_printSoln:Boolean): void {
            solver.printResults(flag_printSoln);
        }
    }
    
    
    // a boolean array to keep track of which place(s) successfully solve the instance
    static val placeSuccessful <: Rail[Boolean] = new Rail[Boolean](Place.MAX_PLACES, false);    

    // main method
    public static def main (args:Array[String](1)) {
        Console.OUT.println("c ==== SatX10 Parallel SAT Solver 1.0 (release 20121215) ====");
        // Store start time
		val lStartTime = System.nanoTime();
		
        // parse command line
        if (args.size != 4 && args.size != 5) {
            Console.ERR.println("USAGE: SatX10 <SatX10ConfigFileName.ppconfig> "
                    + "<maxLenSharedClauses> <outgoingClsBufSize> <filename> [0/1: print solution]");
            return;
        }
        val ppConfigFileName       = args(0);
        val max_len_shared_clauses = Int.parse(args(1));
        val outgoing_cls_buf_size  = Int.parse(args(2));
        val filename               = args(3);
        val flag_printSoln         = (args.size == 4 || !args(4).equals("0"));   // default true: print solution if one is found
        Console.OUT.println("c SatX10 parameters:");
        Console.OUT.println("c   CNF filename              : " + filename);
        Console.OUT.println("c   ppconfig file             : " + ppConfigFileName);
        Console.OUT.println("c   maxLengthSharedClauses    : " + max_len_shared_clauses); 
        Console.OUT.println("c   outgoingClausesBufSize    : " + outgoing_cls_buf_size);
        Console.OUT.println("c   printSolution             : " + flag_printSoln);
        
        // parse .ppconfig file to get the list of solvers to run and 
        // their associated commandline arguments
        val configuration = ExplicitPPConfiguration.makeConfig(filename, 
               max_len_shared_clauses, outgoing_cls_buf_size, ppConfigFileName); 
        
        // create a solver at each place
        val everyone = Dist.makeUnique();
        val s_handle = PlaceLocalHandle.make[RestartableSolver](everyone, 
                () => new RestartableSolver(configuration) );

        // place ID of the master process that handles restart sequence, if 
        // that option is used by the solvers; this place also runs a regular 
        // solver like the rest of the places, responding to getNewRestart()
        // requests whenever it handles incoming messages
        val master_id = here.id();
        Console.OUT.println("c " + here + ": starting master process for restart sequence");

        // launch a solver at each place; wait for all solvers to finish 
        // (by solving or by being killed)
        finish ateach (i in everyone) {
	    try {
		val s = s_handle();

		// define callback to send learned clauses to every other place
		val not_here = new Rail[Place](Place.MAX_PLACES-1, 
					       (i:Int)=>Place.place(i<here.id ? i : i+1));
		val receivers = s.receivers();
		val cb1 = (clauses:Rail[Int]) => {
		    for (j in receivers.values()) at (j) async {
			val local_solver = s_handle();
			//Console.OUT.println("c clauses from afar");
			local_solver.bufferIncomingClauses(clauses);
		    }
		};

		// define callback to the master process for getting restart lengths
		var cb2: () => Int;
		if (here.id == master_id)
		    cb2 = () => s_handle().getNewRestart();
		else
		    cb2 = () => at (Place.place(master_id)) s_handle().getNewRestart();

		// start the solver
		Console.OUT.println("c " + here + ": starting solver");
		val myStartTime = System.nanoTime();
		val result = s.solve(cb1, cb2);
		
		// process the results
		if (result != 0) {   // 0: unsolved, 1: satisfiable; -1: unsatisfiable
		    // the instance was solved here!
		    Console.OUT.println("c " + here + ": INSTANCE SOLVED!");
	        // compute wall-clock time so far and print it
	        val myEndTime   = System.nanoTime();
	        val dSeconds = (myEndTime - myStartTime) / 1000000000.0;
	        Console.OUT.println("c Solver Wall-Clock Time       : " + dSeconds + " seconds");      
		    // send kill signal to all other solvers
		    for (j in not_here.values()) 
			at (j) async s_handle().kill();
		    // record that this place succeeded in solving the instance
		    if (here.id == master_id)
			placeSuccessful(master_id) = true;
		    else {
			val myID = here.id;
			at (Place.place(master_id)) placeSuccessful(myID) = true;
		    }
		}
		else {
		    // the instance wasn't solved; got here most likely because the solver was
		    // killed (or timed out or crashed internally)
		    Console.OUT.println("c " + here + ": aborting");
		}
	    } catch (z:Exception) {
		Console.OUT.println("c " + here + ": aborted because of " + z.toString());
	    }
	    
        } // finish
        
        // scan the placeSuccessful array and for the first solver in it that
        // solved the instance, print final outcome and possibly the solution
        var j : Int = 0;
        while (j < Place.MAX_PLACES && placeSuccessful(j) == false)
            j++;
        if (j < Place.MAX_PLACES)
            // j is the lowest numbered place that was successful; print its results and stats
            at (Place.place(j)) s_handle().printResults(flag_printSoln);
        else
            // no one succeeded; call printResults of the solver at master_id to print 's UNKNOWN' and other stats
            s_handle().printResults(false);
        
        // Get End Time
        val lEndTime   = System.nanoTime();
        // Compute duration in seconds
        val dSeconds = (lEndTime - lStartTime) / 1000000000.0;
        // Output wall clock time 
        Console.OUT.println("c Total time till full exit   : " + dSeconds + " seconds");      
    }
}


// vim: shiftwidth=4:tabstop=4:expandtab
