// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// SatX10
//
// (C) Copyright IBM Corporation 2012.  All Rights Reserved.
//
// Authors: Bard Bloom, David Grove, Benjamin Herta,
//          Vijay Sarawat, Ashish Sabharwal, Horst Samulowitz
// -----------------------------------------------------------------


/* Classes implementing restart sequences for a solver
	// (e.g., SAT solver, CSP solver, etc.) and providing the
	// getNext() method which returns the next "run length"
	// for the restart strategy being implemented.
 * */

public class Seq {
	
	// Abstract base class underlying all other classes
	abstract static class RestartSeq {
		protected var runlength:Int;        // next runlength; starts out as 'uninitialized'
		static val RUNLENGTH_UNASSIGNED:Int= -1;
		public def this(){
			this.runlength = RUNLENGTH_UNASSIGNED;
		}
		
		// pure virtual method, to be implemented by inheriting classes
		abstract protected def getNext():Int;
		
	};

	// Luby Sequence
	public static class LubySeq extends RestartSeq {
		protected val multiplier:Int; // the "scaling factor" with which traditional Luby seq is multiplied
		
		// internal variable to keep track of the number of calls to getNext()
		private var restartsSoFar:Int;
		public def this() { this(1);}
		public def this(multiplier_:Int) {
			this.multiplier=multiplier_;
			this.restartsSoFar = 0;
		}
		public def getNext():Int {
			
				if (runlength == RUNLENGTH_UNASSIGNED)
					runlength = multiplier;
				else {
					var size:Int; var seq:Int;
					var index:Int = restartsSoFar;
					for (size = 1, seq = 0; size < index+1; seq++, size = 2*size+1);

					while (size-1 != index){
						size = (size-1)>>1;
					seq--;
					index = index % size;
					}

					runlength = multiplier * (1 << seq);
				}

				++restartsSoFar;
				return runlength;
			
		}
	}

	// Geometric Sequence
	public static class GeometricSeq extends RestartSeq {
		// using double rather than int to store runlength, as the multiplier is double
		protected var fRunlength:Double;  
		var initialValue:Int;   // initial value; effectively scales the geometric sequence
		var increment: Double ;      // the multiplicative factor applied in each round
		
		public def this() { this(1, 1.2D);}
		public def this(_increment:Double) { this(1, _increment);}
		public def this(_initialValue:Int) { this(_initialValue, 1.2);}
		public def this(_initialValue:Int, _increment:Double) {
			fRunlength = runlength as Double;
			initialValue = _initialValue;
			increment = _increment;
		}
		public def getNext() {
			if (fRunlength == RUNLENGTH_UNASSIGNED as Double)
				fRunlength = initialValue as Double;
			else
				fRunlength *= increment;

			return fRunlength as Int;
		}
	}
	
	// The restart sequence used by Minisat 2 (from 2006 or so)
	//  -> geometric with initial value 100 and increment factor 1.5
	public static class Minisat2Seq extends GeometricSeq {
		public def this() { super(100, 1.5); }
	}

	// The restart sequence used by Minisat 2.2.0 (from 2008)
	//  -> Luby scaled up with a multiplicative factor of 100
	public static class Minisat220Seq extends LubySeq {
		public def this() { super(100); }
	}

}
